#!/usr/bin/perl -w
# This software/database is "United States Government Work" under the terms of
# the United States Copyright Act.  It was written as part of the authors'
# official duties for the United States Government and thus cannot be
# copyrighted.  This software/database is freely available to the public for
# use without a copyright notice.  Restrictions cannot be placed on its present
# or future use. 
# 
# Although all reasonable efforts have been taken to ensure the accuracy and
# reliability of the software and data, the National Human Genome Research
# Institute (NHGRI) and the U.S. Government does not and cannot warrant the
# performance or results that may be obtained by using this software or data.
# NHGRI and the U.S.  Government disclaims all warranties as to performance,
# merchantability or fitness for any particular purpose. 
# 
# In any work or product derived from this material, proper attribution of the
# authors as the source of the software or data should be made, using "NHGRI
# Genome Technology Branch" as the citation. 


# This software is under developments to convert cDNA/mRNA mutation positions to genomic coordinates.
use strict;
use XML::Simple;

my $configFile		= $ARGV[0]; 
my $cfg				= XMLin ($configFile) or die "Usage: perl $0 configuration_file\n";
my $genomicsOutFile	= "$cfg->{user_input}->{mutation_file}\.genomics.output";
my $skippyWebFile	= "$cfg->{user_input}->{mutation_file}\.skippy_web.output";
my $skippyCmdFile	= "$cfg->{user_input}->{mutation_file}\.skippy_command.output";
my $cdsStart		= $cfg->{user_input}->{cdsStart};

open (USER_SNV_FILE_IN, "<", "$cfg->{user_input}->{mutation_file}") or die "Cannot read $cfg->{user_input}->{mutation_file}: $!\n";

# create the header line
my $headerline = join ("\t", qw
~
GenBank_Acc
Variant_ID
Base_change
aa_change
Notes
Genomic_coord
cDNA_coord
Skippy_chromosome
Skippy_coord
Skippy_base_change_web
Skippy_base_change_command
~);
open (GENOMICSOUTFILE,  ">", $genomicsOutFile) or die "Cannot write $genomicsOutFile: $!\n";
open (SKIPPYWEBOUTFILE,  ">", $skippyWebFile) or die "Cannot write $skippyWebFile: $!\n";
open (SKIPPYCMDOUTFILE,  ">", $skippyCmdFile) or die "Cannot write $skippyCmdFile: $!\n";

print GENOMICSOUTFILE "$headerline\n";

my @user_SNV_lns = <USER_SNV_FILE_IN>;
chomp @user_SNV_lns;
close USER_SNV_FILE_IN;

my $coordinateReference;
for (my $idx=1; $idx<=$#user_SNV_lns; $idx++) {

	print "processing >>>$user_SNV_lns[$idx]<<<\n";
	my $muta = parseUserInputLine ($user_SNV_lns[$idx]);

	$muta->{user_id}			= (defined $muta->{user_id})	? $muta->{user_id}	: '';
	$muta->{user_reported_exon}	= (defined $muta->{user_reported_exon})	? $muta->{user_reported_exon}	: '';
	$muta->{note}				= (defined $muta->{note})		? $muta->{note}		: $user_SNV_lns[$idx];
	$muta->{note}				=~ s/\t/\|/g;
	my $output_line = "$muta->{acc}\t$muta->{user_id}\t$muta->{bic_fashion_base_chg}\t$muta->{AA_change}\t$muta->{note}";

		my $ret = getGenomicsCoordinates ($muta);

		if (($coordinateReference->{strand} eq '-') and (defined $ret->{start})) {
			my $tmp					= $ret->{endExon};
			$ret->{endExon}			= $ret->{startExon};
			$ret->{startExon}		= $tmp;
			$tmp					= $ret->{endCoordinate};
			$ret->{endCoordinate}	= $ret->{startCoordinate};
			$ret->{startCoordinate}	= $tmp;
		}
		
		$output_line .= (defined $ret->{startCoordinate}) ? "\t$ret->{startCoordinate}" : "\t";
		$output_line .= (defined $ret->{start}) ? "\t$ret->{start}" : "\t";

		# create SNHGRI SKIPPY site's required columns
		my $Skippy_chromosome	= $coordinateReference->{chrom}; 
		my $Skippy_coordinate	= (defined $ret->{startCoordinate}) ? $ret->{startCoordinate} : '';
		#my $Skippy_base_change	= (defined $ret->{bic_fashion_base_chg}) ? $ret->{bic_fashion_base_chg} : '';
		my $Skippy_base_chg_web	= (defined $ret->{bic_fashion_base_chg}) ? $ret->{bic_fashion_base_chg} : '';
		my $Skippy_base_chg_cmd	= (defined $ret->{bic_fashion_base_chg}) ? $ret->{bic_fashion_base_chg} : '';
		$Skippy_base_chg_web	=~ s/ to /\-\>/;
		$Skippy_base_chg_cmd	=~ s/ to / /;
		$output_line			.= "\t$Skippy_chromosome\t$Skippy_coordinate\t$Skippy_base_chg_web\t$Skippy_base_chg_cmd";

	print GENOMICSOUTFILE "$output_line\n";
	print SKIPPYWEBOUTFILE "$Skippy_chromosome\t$Skippy_coordinate\t$Skippy_base_chg_web\n";
	print SKIPPYCMDOUTFILE "$Skippy_chromosome\t$Skippy_coordinate\t$Skippy_base_chg_cmd\n";
}

print "\ndone!\n";
close GENOMICSOUTFILE;
close SKIPPYWEBOUTFILE;
close SKIPPYCMDOUTFILE;


# ret->acc					: RefSeq or GenBank acc
# ret->ver					: version for gbStatus
# ret->start				: start position in the transcript
# ret->end					: end position in the transcript
# ret->intronic				: flag for intronic or not
# ret->offset_calculator	: plus or minus to calculate if intronic flag is true
# revised for mRNA2genomics.pl, some of above are preserved
# ret->user_id				: User defined accession
# ret->pos_within_codon		: position within the reported codon, e.g., 1 is the first nucleotide
# ret->user_reported_seq	: user reported nucleotide sequence
# ret->bic_fashion_base_chg	: NHGRI BIC fashion nucleotide base change
# ret->AA_change			: NHGRI BIC fashion Amino Acid change of that codon
sub parseUserInputLine {
	my $ln = shift;
	my ($gb_acc, $gb_ver, $muta_start, $muta_end, $question_mark, $muta_type);
	my $ret = {};

	# all user input has to be tab delimited.
	$ln = _trim ($ln);
	my @tokens = split (/\t/, $ln);

	if ((uc $cfg->{user_input}->{nomenclature_fashion} eq 'MRNA2GENOMICS') and defined $tokens[3] and $tokens[3] !~ /\D/) {
		# this input is currently for single nucleotide only
		# user input file has
		# User_ID  \t Base_change \t AA_change \t Codon_number [\t additional notes]
		# ID00001  \t aATG-GTG    \t Met-Val   \t 1            [\t Breast and/or ovarian cancer]
		$ret->{user_id}		= $tokens[0];
		# first one is the accession with version
		my @acc_parts = split (/\./, $cfg->{references}->{accession_number});
		if (scalar @acc_parts > 1) {
			$ret->{ver} = $acc_parts[$#acc_parts];
			$ret->{acc} = join (".", ($acc_parts[0] .. $acc_parts[$#acc_parts-1]));
		}
		if (scalar @acc_parts > 0) {
			$ret->{acc} = $acc_parts[0];
		}
		else {
			$ret->{acc} = $tokens[0];
		}
		
		$ret->{pos_within_codon} = 0;
		if ($tokens[1] =~ /[a-z]*([A-Z]+)[a-z]*\-[a-z]*([A-Z]+)/ ) {
			my $pre			= $1;
			my $post		= $2;
			while (substr($pre,0,1) eq substr($post,0,1)) {
				$pre		= substr ($pre, 1);
				$post		= substr ($post,1);
				$ret->{pos_within_codon}++;
			}
			$ret->{pos_within_codon}++;
			$ret->{user_reported_seq}		= uc substr($pre,0,1);
			$ret->{bic_fashion_base_chg}	= substr($pre,0,1)." to ".substr($post,0,1);
		}
		$ret->{AA_change}	= $tokens[2];
		$ret->{start}		= ($tokens[3] - 1) * 3 + $ret->{pos_within_codon} + $cfg->{user_input}->{cdsStart} - 1;
		$ret->{end}			= $ret->{start}; 
	}
	elsif ((uc $cfg->{user_input}->{nomenclature_fashion} eq 'BIC') and defined $tokens[3] and defined $tokens[4] and $tokens[3] !~ /\D/ and $tokens[4] !~ /\D/) {
		# user input file has
		# GenBank  \t (bic) ACC_NUMBER \t (bic) EXON \t NT     \t BASE_CHANGE \t AA_CHANGE \t DESIGNATION
		# U14680.1 \t 3384             \t I-11       \t 4215+1 \t G to A      \t -         \t IVS11+1G>A
		# U14680.1 \t 5192             \t I-11       \t 4215+1 \t G to A      \t -         \t IVS11+1G>A
		# for now, single nucleotide mutations first
		# 4th token is mutation location (nt), offset from transcript start
		# first one is the accession with version
		my @acc_parts = split (/\./, $tokens[0]);
		if (scalar @acc_parts > 1) {
			$ret->{ver} = $acc_parts[$#acc_parts];
			$ret->{acc} = join (".", ($acc_parts[0] .. $acc_parts[$#acc_parts-1]));
		}
		else {
			$ret->{acc} = $tokens[0];
		}
		$ret->{user_reported_exon} = $tokens[2];
		$ret = parseBICMutationStart ($ret, $tokens[3]);
		# 5th token is the base_change and have range info
		$ret = parseBICMutationEnd ($ret, $tokens[4]);
	}
	elsif ((uc $cfg->{user_input}->{nomenclature_fashion} eq 'HGVS') and defined $tokens[1] and $tokens[1] =~ /\d/) {
		# first one is the accession with version
		my @acc_parts = split (/\./, $tokens[0]);
		if (scalar @acc_parts > 1) {
			$ret->{ver} = $acc_parts[$#acc_parts];
			$ret->{acc} = join (".", ($acc_parts[0] .. $acc_parts[$#acc_parts-1]));
		}
		else {
			$ret->{acc} = $tokens[0];
		}
		($ret->{start}, $ret->{end}, $ret->{question_mark}) = parseMutationPositions ($tokens[1]);
	}

	return $ret;
}


# to find part of muta hash by returning
# ret->start
# ret->intronic
# ret->offset_calculator
# ret->offset
sub parseBICMutationStart {
	my ($ret, $muta_str) = @_;
	
	if ($muta_str =~ /(\d+)(\D)(\d+)/) {
		# intronic mutation
		$ret->{start} = $1;
		$ret->{operand} = $2;
		$ret->{offset} = $3;
		
		# necessary adjustment for NHGRI BIC site's intron/exon boundaries when using BIC data
		if ((uc $ret->{acc} eq 'U14680') and defined $ret->{operand}) {
			$ret->{start} = 667		if ($ret->{start} == 666);
			$ret->{start} = 790		if ($ret->{start} == 789);
			$ret->{start} = 4305	if ($ret->{start} == 4303);
			$ret->{start} = 4604	if ($ret->{start} == 4564);
			$ret->{start} = 5106	if ($ret->{start} == 5006);
			$ret->{start} = 5272	if ($ret->{start} == 5274);
			$ret->{start} = 5313	if ($ret->{start} == 5311);
			$ret->{start} = 5526	if ($ret->{start} == 5527);
			$ret->{start} = 666		if ($ret->{start} == 665);
			$ret->{start} = 789		if ($ret->{start} == 788);
			$ret->{start} = 4304	if ($ret->{start} == 4302);
			$ret->{start} = 4603	if ($ret->{start} == 4563);
			$ret->{start} = 5105	if ($ret->{start} == 5005);
			$ret->{start} = 5271	if ($ret->{start} == 5273);
			$ret->{start} = 5312	if ($ret->{start} == 5310);
			$ret->{start} = 5525	if ($ret->{start} == 5526);
		}
		elsif ((uc $ret->{acc} eq 'U43746') and defined $ret->{operand}) {
			$ret->{start} = 7845	if ($ret->{start} == 7829);
			$ret->{start} = 7846	if ($ret->{start} == 7830);
			$ret->{start} = 10926	if ($ret->{start} == 10836);
		}
	}
	elsif ($muta_str !~ /\D/) {
		# exonic mutation
		if ($muta_str =~ /(\d+)/) {
			$ret->{start} = $1;
		}
		delete $ret->{operand};
		delete $ret->{offset};
	}
	elsif ($muta_str !~ /(\d+)/) {
		print "!!! Currently I cannot process this type of mutations !!!\n";
		delete $ret->{start};
		delete $ret->{operand};
		delete $ret->{offset};
	}
	
	return $ret;
}


sub parseBICMutationEnd {
	my ($ret, $str) = @_;
	return $ret unless (defined $ret->{start});
	
	if ($str =~ /(del|ins)\s+([a|t|g|c|u]+)\s*$/i ) {
		# del with cDNA notes
		$ret->{mutationType} = uc $1;
		$ret->{cDNAseq} = $2;
		$ret->{end} = $ret->{start} + length($ret->{cDNAseq}) - 1;
	}
	elsif ($str =~ /(del|ins)\s+(\d+)\s*$/i ) {
		$ret->{mutationType} = uc $1;
		$ret->{end} = $ret->{start} + $2 - 1;
	}
	elsif ($str =~ /(del|ins)\s+(\d+)\s+(del|ins)\s+(\d+)/i) {
		$ret->{mutationType} = uc "$1/$3";
		$ret->{end} = ($2 > $4)
			? $ret->{start} + $2 - 1
			: $ret->{start} + $4 - 1;
	}
	elsif ($str =~ /([a|t|g|c|u]+)\s+to\s+([a|t|g|c|u]+)/i) {
		$ret->{end} = (length($1) == length($2))
			? $ret->{start} + length($1) - 1
			: (length($1) > length($2))
				? $ret->{start} + length($1) - 1
				: $ret->{start} + length($2) - 1;
	}

	if ((uc $ret->{acc} eq 'U14680') and defined $ret->{operand}) {
		$ret->{end} -= 1	if ($ret->{start} == 667);
		$ret->{end} -= 1	if ($ret->{start} == 790);
		$ret->{end} -= 2	if ($ret->{start} == 4305);
		$ret->{end} -= 40	if ($ret->{start} == 4604);
		$ret->{end} -= 100	if ($ret->{start} == 5106);
		$ret->{end} += 2	if ($ret->{start} == 5272);
		$ret->{end} -= 2	if ($ret->{start} == 5313);
		$ret->{end} += 1	if ($ret->{start} == 5526);
		$ret->{end} += 1	if ($ret->{start} == 666);
		$ret->{end} += 1	if ($ret->{start} == 789);
		$ret->{end} += 2	if ($ret->{start} == 4304);
		$ret->{end} += 40	if ($ret->{start} == 4603);
		$ret->{end} += 100	if ($ret->{start} == 5105);
		$ret->{end} -= 2	if ($ret->{start} == 5271);
		$ret->{end} += 2	if ($ret->{start} == 5312);
		$ret->{end} -= 1	if ($ret->{start} == 5525);
	}
	elsif ((uc $ret->{acc} eq 'U43746') and defined $ret->{operand}) {
		$ret->{end} += 16	if ($ret->{start} == 7829);
		$ret->{end} += 16	if ($ret->{start} == 7830);
		$ret->{end} += 90	if ($ret->{start} == 10836);
	}

	return $ret;
}


# to work with hgvs nomenclature and bic nomenclature, we will test the flag and call proper parsing.
# we will have following basic patterns:
# minue sign after exon's boundary: how many nucleotides go back into intron from exon starting position
# plus sign after exon's boundary: how many nucleotides go forward into intron from exon ending position
# "del" after a position X, then followed with number Y: Y nucleotides are deleted from coding position X
# "ins" after a position X, then followed with number Y: Y nucleotides are inserted from coding position X.
sub parseMutationPositions {
	my $mutaString = shift;
	my ($muta_start, $muta_end, $question_mark, $muta_type);

	$question_mark = ($mutaString =~ /\?/) ? 1 : 0;
	
	my @tokens = split (/\_/, $mutaString);

	$tokens[0] = _trim ($tokens[0]);
	if ($tokens[0] =~ /^c\.\-/) {
		# hgvs naming allows ommission of exon 1 nucleotide 1 and just put the offset from that, a short hand notation
		$muta_start = 1;
	}
	elsif ($tokens[0] =~ /[\-|\+|\D](\d+)/) {
		$muta_start = $1;
	}

	if (scalar @tokens > 1) {
		if ($tokens[1] =~ /[\-|\+\D]*(\d+)/) {
			$muta_end = $1;
		}
	}
	else {
		$muta_end = $muta_start;
	}
	
	return ($muta_start, $muta_end, $question_mark);
}


sub getGenomicsCoordinates {
	my $muta = shift;
	my $ret_info = {};

	return $muta unless (
		defined $muta->{start} and defined $muta->{end} 
		and $muta->{start} =~ /\d+/ and $muta->{start} !~ /\D/ 
		and $muta->{end} =~ /\d+/ and $muta->{end} !~ /\D/);

	if (defined $coordinateReference) {
		if ($coordinateReference->{acc} eq $muta->{acc}) {
			$muta = getMutationStartCoordinate ($muta);
		}
		else {
			$coordinateReference = findCoordinatesReference ($muta);
			if ($coordinateReference->{acc} eq $muta->{acc}) {
				$muta = getMutationStartCoordinate ($muta);
			}
		}
	}
	else {
		$coordinateReference = findCoordinatesReference ($muta);
		if ($coordinateReference->{acc} eq $muta->{acc}) {
			$muta = getMutationStartCoordinate ($muta);
		}
	}

	return $muta;
}


# refGene.txt format:
##bin	name	chrom	strand	txStart	txEnd	cdsStart	cdsEnd	exonCount	exonStarts	exonEnds	score	name2	cdsStartStat	cdsEndStat	exonFrames
#104	NM_000059	chr13	+	32889616	32973809	32890597	32972907	27	32889616,32890558,32893213,32899212,32900237,32900378,32900635,32903579,32905055,32906408,32910401,32918694,32920963,32928997,32930564,32931878,32936659,32937315,32944538,32945092,32950806,32953453,32953886,32954143,32968825,32971034,32972298,	32889804,32890664,32893462,32899321,32900287,32900419,32900750,32903629,32905167,32907524,32915333,32918790,32921033,32929425,32930746,32932066,32936830,32937670,32944694,32945237,32950928,32953652,32954050,32954282,32969070,32971181,32973809,	0	BRCA2	cmpl	cmpl	-1,0,1,1,2,1,0,1,0,1,1,1,1,2,1,0,2,2,0,0,1,0,1,0,1,0,0,
# convert all coordinates to one based
sub parseRefGeneLine {
	my $ln = _trim (shift);
	my $ref = {};
	
	my @tokens = split (/\t/, $ln);
	$ref->{acc}			= $tokens[1];								# name
	$ref->{chrom}		= $tokens[2];
	$ref->{strand}		= $tokens[3];
	$ref->{txStart}		= $tokens[4]+1;
	$ref->{txEnd}		= $tokens[5];
	$ref->{cdsStart}	= $tokens[6]+1;
	$ref->{cdsEnd}		= $tokens[7];
	$ref->{exonCount}	= $tokens[8];
	my @exonStarts		= split (/,/, $tokens[9]);
	my @exonEnds		= split (/,/, $tokens[10]);
	my @exonLengths;
	for (my $i=0; $i < scalar (@exonStarts); $i++) {
		push @exonLengths, ($exonEnds[$i] - $exonStarts[$i]); 
		$exonStarts[$i]++;
	}
	$ref->{exonStarts}	= \@exonStarts;
	$ref->{exonEnds}	= \@exonEnds;
	$ref->{exonLengths}	= \@exonLengths;

	return $ref;
}


# U14680_all_mrna_hg19.txt has no exon 4
# since this is minus strand, the exons are backword
# example here is revised to match BIC database exon/introns. Original UCSC downloaded annotations are 24 coding blocks
##bin	matches	misMatches	repMatches	nCount	qNumInsert	qBaseInsert	tNumInsert	tBaseInsert	strand	qName	qSize	qStart	qEnd	tName	tSize	tStart	tEnd	blockCount	blockSizes	qStarts	tStarts
#899	5703	0	0	1	1	7	23	73988	-	U14680	5711	0	5711	chr17	81195210	41197694	41277387	23	125,61,74,55,84,41,78,88,311,191,127,172,89,3426,77,46,106,140,89,78,54,99,100	0,125,186,260,315,399,440,518,606,917,1108,1235,1407,1496,4922,4999,5045,5151,5291,5380,5458,5512,5611,	41197694,41199659,41201137,41203079,41209068,41215349,41215890,41219624,41222944,41226347,41228504,41234420,41242960,41243451,41247862,41249260,41251791,41256138,41256884,41258472,41267742,41276033,41277287,
sub parseAllmRNALine {
	my $ln = _trim (shift);
	my $ref = {};
		
	my @tokens = split (/\t/, $ln);

	$ref->{acc}			= $tokens[10];								# name
	$ref->{chrom}		= $tokens[14];
	$ref->{strand}		= $tokens[9];
	$ref->{txStart}		= $tokens[16]+1;							# make it 1 based
	$ref->{txEnd}		= $tokens[17];

	$ref->{cdsStart}	= $tokens[16]+1;							# make it 1 based
	$ref->{cdsEnd}		= $tokens[17];

	$ref->{exonCount}	= $tokens[18];
	my @exonStarts		= split (/,/, $tokens[21]);
	my @blockSizes		= split (/,/, $tokens[19]);
	my @exonEnds;
	for (my $i=0; $i < scalar (@exonStarts); $i++) {
		$exonEnds[$i] = $exonStarts[$i] + $blockSizes[$i];
		$exonStarts[$i]++;											# make it 1 based
	}	
	$ref->{exonStarts}	= \@exonStarts;
	$ref->{exonEnds}	= \@exonEnds;
	$ref->{exonLengths}	= \@blockSizes;

	return $ref;
}


sub findCoordinatesReference {
	my $muta = shift;

	my ($cmd, @ret, @filtered_ret);
	$cmd = qq~grep $muta->{acc} $cfg->{references}->{coordinates_file}~;
	@ret = `$cmd`;
	if (uc $cfg->{references}->{accession_file_type} ne 'GENBANK') {
		# RefSeq type annotations from UCSC
		# second column is the accession name
		foreach my $retln (@ret) {
			my @tmp = split /\t/, $retln;
			push @filtered_ret, $retln if ($tmp[1] eq $muta->{acc});
		}
		if (scalar @filtered_ret > 0) {
			$coordinateReference = parseRefGeneLine ($filtered_ret[0]);
		}
		else {
			die "$0 cannot find coordinate reference: $cfg->{references}->{coordinates_file} does not contain $muta->{acc}!\n";
		}
	}
	else {
		# all_mrna type annotations from UCSC
		# eleventh column is the accession name
		foreach my $retln (@ret) {
			my @tmp = split /\t/, $retln;
			push @filtered_ret, $retln if ($tmp[10] eq $muta->{acc});
		}
		if (scalar @filtered_ret > 0) {
			$coordinateReference = parseAllmRNALine ($filtered_ret[0]);
		}
		else {
			die "$0 cannot find coordinate reference: $cfg->{references}->{coordinates_file} does not contain $muta->{acc}!\n";
		}
	}
	
	return $coordinateReference;
}


sub getMutationStartCoordinate {
	my $muta = shift;

	if (uc $cfg->{references}->{accession_file_type} ne 'GENBANK') {
		$muta->{start}	= $muta->{start} - $cfg->{user_input}->{cdsStart} + 1;
		$muta->{end}	= $muta->{end} - $cfg->{user_input}->{cdsStart} + 1;
	}
	if ($coordinateReference->{strand} eq '+') {
		for (my $i=1; $i<=(scalar @{$coordinateReference->{exonStarts}}); $i++) {

			# go next exon if this exon is not in of coding sequence
			next if ($coordinateReference->{exonEnds}->[$i-1] < $coordinateReference->{cdsStart});

			# if cdsStart is within this exon
			if (!defined $muta->{startCoordinate} 
			and $coordinateReference->{exonStarts}->[$i-1] <= $coordinateReference->{cdsStart}
			and $coordinateReference->{exonEnds}->[$i-1] >= $coordinateReference->{cdsStart}) {
				
				my $currCodingLen = 0;
				# subtract coding sequences length until the start of mutation falls within an exon start and end
				while ($currCodingLen < $muta->{start}) {
					$currCodingLen += ($currCodingLen == 0)
						? $coordinateReference->{exonEnds}->[$i-1] - $coordinateReference->{cdsStart} + 1
						: $coordinateReference->{exonLengths}->[$i-1];
					$i++;
				}
				$muta->{startExon}			= $i-1;
				$muta->{startCoordinate}	= $coordinateReference->{exonStarts}->[$i-2] - 1 
											+ $coordinateReference->{exonLengths}->[$i-2] 
											- $currCodingLen + $muta->{start};
				# subtract coding sequences length until the end of mutation falls within an exon start and end
				while ($currCodingLen < $muta->{end}) {
					$currCodingLen += ($currCodingLen == 0)
						? $coordinateReference->{exonEnds}->[$i-1] - $coordinateReference->{cdsStart} + 1
						: $coordinateReference->{exonLengths}->[$i-1];
					$i++;
				}
				$muta->{endExon}			= $i-1;
				$muta->{endCoordinate}		= $coordinateReference->{exonStarts}->[$i-2] - 1 
											+ $coordinateReference->{exonLengths}->[$i-2] 
											- $currCodingLen + $muta->{end};
			}
		}
		# adjust if this is an intronic mutation
		if (defined $muta->{operand} and $muta->{operand} eq '-') {
			$muta->{startCoordinate} -= $muta->{offset};
			$muta->{endCoordinate}	 -= $muta->{offset};
		}
		if (defined $muta->{operand} and $muta->{operand} eq '+') {
			$muta->{startCoordinate} += $muta->{offset};
			$muta->{endCoordinate}	 += $muta->{offset};
		}
	}
	else {

		# count down fashion due to reversed strand
		for (my $i=(scalar @{$coordinateReference->{exonStarts}}); $i>0 and defined $muta->{start}; $i--) {

			# go next exon if this exon is not in of coding sequence
			next if ($coordinateReference->{exonStarts}->[$i-1] > $coordinateReference->{cdsEnd});

			# if cdsEnd is within this exon
			if (!defined $muta->{startCoordinate} 
			and $coordinateReference->{exonStarts}->[$i-1] <= $coordinateReference->{cdsEnd}
			and $coordinateReference->{exonEnds}->[$i-1] >= $coordinateReference->{cdsEnd}) {
				
				my $currCodingLen = 0;
				# subtract coding sequences length until the start of mutation falls within an exon start and end
				while ($currCodingLen < $muta->{start}) {
					$currCodingLen += ($currCodingLen == 0)
						? $coordinateReference->{cdsEnd} - $coordinateReference->{exonStarts}->[$i-1] + 1
						: $coordinateReference->{exonLengths}->[$i-1];
					$i--;
				}
				# the start of the mutation is within this exon
				$muta->{startCoordinate} = $coordinateReference->{exonStarts}->[$i] + $currCodingLen - $muta->{start};
				$muta->{startExon} = (scalar @{$coordinateReference->{exonStarts}}) - $i;

				# subtract coding sequences length until the end of mutation falls within an exon start and end
				while ($currCodingLen < $muta->{end}) {
					$currCodingLen += ($currCodingLen == 0)
						? $coordinateReference->{cdsEnd} - $coordinateReference->{exonStarts}->[$i-1] + 1
						: $coordinateReference->{exonLengths}->[$i-1];
					$i--;
				}

				# the start of the mutation is within this exon
				$muta->{endCoordinate} = $coordinateReference->{exonStarts}->[$i] + $currCodingLen - $muta->{end};
				$muta->{endExon} = (scalar @{$coordinateReference->{exonStarts}}) - $i;
			}
		}

		# adjust if this is an intronic mutation
		if (defined $muta->{operand} and $muta->{operand} eq '-') {
			$muta->{startCoordinate} += $muta->{offset};
			$muta->{endCoordinate}	 += $muta->{offset};
		}
		if (defined $muta->{operand} and $muta->{operand} eq '+') {
			$muta->{startCoordinate} -= $muta->{offset};
			$muta->{endCoordinate}	 -= $muta->{offset};
		}
	}
	
	if (uc $cfg->{references}->{accession_file_type} ne 'GENBANK') {
		$muta->{start}	= $muta->{start} + $cfg->{user_input}->{cdsStart} - 1;
		$muta->{end}	= $muta->{end} + $cfg->{user_input}->{cdsStart} - 1;
	}

	return $muta;
}


sub _trim {
	my $arg = shift;
	$arg =~ s/^\s*//;
	$arg =~ s/\s*$//;
	return $arg
}
